import {
  json,
  select,
  timeParse,
  colorScale,
  hierarchy,
  zoom,
} from 'd3';
import { compactEventSequences } from './compactEventSequences';
import { attributeHistogram } from './attributeHistogram';
import { eventSequences } from './eventSequences';
import { sequenceInfo } from './sequenceInfo';






// -------- Visualization Constants -------- //
const leftSvg = select('#left-svg');
const rightSvg = select('#right-svg');
const middleSvg = select('#middle-svg');
const leftHeight = leftSvg.attr('height');
const rightHeight = rightSvg.attr('height');
const middleHeight = middleSvg.attr('height');
const leftWidth = document.getElementById('left-div').offsetWidth;
const rightWidth = document.getElementById('right-div').offsetWidth;
const middleWidth = document.getElementById('middle-div').offsetWidth;

// Widget Groups
const cesG = middleSvg.append('g');
const sequencesG = middleSvg.append('g');






// -------- State and Handlers -------- //
let data = [];
let regex;

let selectedEvents = {};
let highlightedSequence;
let highlightedSequenceId;

let eventsMatched, totalEvents;
let sequencesMatched, totalSequences;


const handleRule = () => {
  const rule = document.getElementById('rule').value;
  
  if(rule === '') {
    regex = null;
  } else {
    try {
      regex = new RegExp(rule);
    } catch(e) {
      console.log('Invalid Regex', e);
      return;
    }
  }
  
  eventsMatched = totalEvents = 0;
  sequencesMatched = totalSequences = 0;
  render();
}

const selectEvent = (id, depth) => {
  if(selectedEvents[id]) {
    selectedEvents[id][depth] = (selectedEvents[id][depth] === undefined)
    														? true
    														: !selectedEvents[id][depth];
  } else {
    selectedEvents[id] = { [depth]: true };
  }
  
  // Calculate Rule
  let rule = '';
  const sequence = getSequence(id);
  const events = Object.keys(selectedEvents[id]);
  events.map(d => +d).forEach((d, i, arr) => {
    const { N, CN, CT } = sequence[d].data.data;
    const newRule = getRule(N, CN, CT);
    rule = rule + newRule;
  });
  
  document.getElementById('rule').value = rule;
  handleRule();
  render();
}

const highlightSequence = (sequence, id) => {
  highlightedSequence = sequence.slice(1);
  highlightedSequenceId = id;
  console.log(id);
  render();
}

const handleEvent = (bool) => {
  if(bool) eventsMatched += 1;
  totalEvents += 1;
  
  // Display Events
  const eventsMeta = document.getElementById('events-meta');
  const num = (eventsMatched / totalEvents) * 100;
  const display = (isNaN(num) || !isFinite(num)) 
  								? 0.00 : num.toFixed(2);
  eventsMeta.innerHTML = `Events Matched: ${display}%`
}

const handleSequence = (bool) => {
  if(bool) sequencesMatched += 1;
  totalSequences += 1;
  
  // Display Sequences
  const sequencesMeta = document.getElementById('sequences-meta');
  const num = (sequencesMatched / totalSequences) * 100;
  const display = (isNaN(num) || !isFinite(num))
                  ? 0.00 : num.toFixed(2);
  sequencesMeta.innerHTML = `Sequences Matched: ${display}%`;
}




// -------- Rendering and Data Loading -------- //
const render = () => {
  
  // cesG.call(compactEventSequences, {
  //   data,
  //   spacing: { right: 0, bottom: 0 },
  //   margin: { top: 20, right: 20, bottom: 20, left: 10 },
  //   length: 1,
  // });
  
  sequencesG.call(eventSequences, {
    data,
    rule,
    regex,
    handleEvent,
    handleSequence,
    selectedEvents,
    highlightSequence,
    highlightedSequence,
    highlightedSequenceId,
    spacing: { right: 6, bottom: 6 },
    margin: { top: 20, right: 20, bottom: 20, left: 40 },
    length: 20,
  });
  
  rightSvg.call(sequenceInfo, {
    regex,
    selectEvent,
    selectedEvents,
  	sequence: highlightedSequence,
    sequenceId: highlightedSequenceId,
  });
  
  
  // Set parent height for scrolling
  const height = middleSvg.node().getBBox().height + 60;
  middleSvg.attr('height', height);

}


json('data.json').then((loadedData) => {
  
  // Attach Event Listeners
  const ruleInput = document.getElementById('rule');
  ruleInput.addEventListener('input', handleRule);
  
  const ruleButton = document.getElementById('calculate');
  ruleButton.addEventListener('click', eventsToRules);
  
  // Get attribute counts
  const dip = (d) => d.data.dest_ip;
  const sip = (d) => d.data.source_ip;
  const dport = (d) => d.data.dest_port;
  const sport = (d) => d.data.source_port;
  
  const dest_ip = getCounts(loadedData, dip);
  const source_ip = getCounts(loadedData, sip);
  const dest_port = getCounts(loadedData, dport);
  const source_port = getCounts(loadedData, sport);
  
  leftSvg.append('g')
  	.call(attributeHistogram, {
      margin: { top: 40, right: 40, bottom: 0, left: 40 },
      width: leftWidth,
      height: 180,
      data: dest_ip,
      index: 'Destination IP',
  	});
  
  leftSvg.append('g')
    .attr('transform', 'translate(0, 220)')
    .call(attributeHistogram, {
      margin: { top: 40, right: 20, bottom: 0, left: 40 },
      width: leftWidth,
      height: 180,
      data: dest_port,
      index: 'Destination Port',
  	});
  
  leftSvg.append('g')
    .attr('transform', 'translate(0, 440)')
    .call(attributeHistogram, {
  		margin: { top: 40, right: 20, bottom: 0, left: 40 },
      width: leftWidth,
      height: 180,
    	data: source_ip,
      index: 'Source IP',
  	});
  
  
  // aggregate unique sequences and insert counts
  let newData = loadedData.reduce((acc, val) => {
    const gui = val.data.gui;
    
    if(acc.hasOwnProperty(gui)) {
      acc[gui].data.count += 1;
    } else {
      val.data['count'] = 1;
      acc[gui] = val;
    }
    
    return acc;
  }, {});
  
  for(let key in newData) data.push(hierarchy(newData[key]).descendants());
  data.sort((a,b) => b[0].data.data.count - a[0].data.data.count)
  for(let i in data) data[i][1].data.data.id = i;
  console.log(data);
  console.log(newData);
  console.log(loadedData);
  
  render();
});








// -------- Helper Functions -------- //
const getCounts = (data, fetch) => {
  
  // calculate counts
  const counts = data.reduce((acc, val) => {
  	const d = fetch(val);
    if(acc.hasOwnProperty(d)) {
      acc[d].value += 1;
    } else {
      acc[d] = {value: 1, name: d};
    }
    
    return acc;
  }, {});
  
  // turn to list
  let countsList = [];
  for(let key in counts) {
    countsList.push(counts[key])
  }
  
  // sort and truncate
  countsList.sort((a,b) => b.value - a.value)
  if(countsList.length > 10) {
    let total = 0;
  	for(let i = 9; i < countsList.length; i++) {
      total += countsList[i].value;
    }
    countsList = countsList.slice(0, 9);
    countsList.push({name: 'REST', value: total})
  }
  
  return countsList;
}


function showRules(rules) {
  const ruleDiv = document.getElementById("rules");
  const ruleRawDiv = document.getElementById("rules_raw");
  
  rules.forEach((val, i, arr) => {
    const id = val.id;
    const content = val.rule;
    const sequence = getSequence(id);
    const sequenceNum = getSequenceNum(id);
    const newElement = document.createElement('span');
    const newContent = document.createTextNode(content);
    
    newElement.appendChild(newContent);
    newElement.addEventListener('click', () => highlightSequence(sequence, sequenceNum));
    
    ruleDiv.appendChild(newElement);
    ruleDiv.appendChild('</br>');
    
  	// const newElement = `<span>${val}</span><br/><br/>`;
  	// ruleHtml = ruleHtml + newElement;
  });
  
  // ruleDiv.innerHTML = ruleHtml;
  ruleRawDiv.innerText = JSON.stringify(rules);
}


function eventsToRules() {
  let rules = [];
  const sequenceKeys = Object.keys(selectedEvents);
  for(let key_s of sequenceKeys) {
    
    let rule = '';
    const id = parseInt(key_s);
    const sequence = getSequence(id);
    const events = Object.keys(selectedEvents[key_s]);
    events.map(d => +d).forEach((d, i, arr) => {
      const { N, CN, CT } = sequence[d].data.data;
    	const newRule = getRule(N, CN, CT);
      rule = rule + newRule;
    });
    
    rules.push({id: id, rule: rule});
  }
  
  showRules(rules);
  console.log(rules);
}

function cleanValue(value) {
  return value && value !== '-' && value !== '*';
}

function getRule(n, cn, ct) {
  let rule = '';
  if(cleanValue(ct)) rule = `.*CT=${ct}`;
  if(cleanValue(cn)) rule = `.*CN=${cn}` + rule;
  if(cleanValue(n)) rule = `.*N=${n}` + rule;
  
  return `.*${rule}.*`;
}

function getSequenceNum(id) {
  for(let i in data) {
    if(data[i][0].data.data.id === id)
      return i;
  }
}

function getSequence(id) {
  for(let sequence of data) {
  	if(sequence[0].data.data.id === id) {
			return sequence;
    }
  }
  return [];
}



